@@ -1,3 +1,8 @@
+import os
+import sys
+
+sys.path.insert(0, os.path.abspath(".."))
+
 project = "python-arango"
 copyright = "2016-2022, Joohwan Oh"
 author = "Joohwan Oh"
